import numpy as np
import matplotlib.pyplot as plt
import os
import zipfile

# ########## TUTORIAL TOOLS ################

def download_tutorial_10(target_file='tutorial_10_files.zip'):
    source_url = f'https://is.muni.cz/el/fi/jaro2024/PA228/um/exercises/tutorial_10/{target_file}'
    print(f'Downloading data from {source_url} to {target_file}')
    if not os.path.exists(target_file):
        print(f'\tDownloading ... {source_url}')
        content = requests.get(source_url).content
        with open(target_file, 'wb') as file:
            file.write(content)
    else:
         print(f'\tFile already downloaded')   
    folder_name = target_file.replace('.zip', '')
    if not os.path.exists(folder_name):
        print(f'\tExtracting to ... {folder_name}')
        with zipfile.ZipFile(target_file, 'r') as zip_ref:
            zip_ref.extractall(folder_name)
    else:
        print(f'\tFile already extracted') 

def load_exercise1(fname):
    with open(fname, 'r') as file:
        lines = file.readlines()
        text = lines[0][:-1]
        words = text.split(' ')
        queries = np.empty((len(words), 3))
        keys = np.empty((len(words), 3))
        values = np.empty((len(words), 3))
        embeddings = {}
        for idx, l in enumerate(lines[1:len(words)+1]):
            queries[idx] = [float(f) for f in l.split(' ')]
        for idx, l in enumerate(lines[len(words)+1:2*len(words)+1]):
            keys[idx] = [float(f) for f in l.split(' ')]
        for idx, l in enumerate(lines[2*len(words)+1:3*len(words)+1]):
            values[idx] = [float(f) for f in l.split(' ')]
            
        for idx, l in enumerate(lines[3*len(words)+1:]):
            word, vals = l.split(':')
            embeddings[word] = [float(f) for f in vals.split(' ')]
    return queries, keys, values, text, embeddings

def embed(words, vocab, vocab_emb):
	emb = []
	for w in words:
		emb.append(vocab_emb[vocab.index(w)])
	return emb

def disembed(emb, vocab, vocab_emb):
	disemb = []
	for e in emb:
		mindist = np.finfo(np.float32).max
		minword = None
		for v, ve in zip(vocab, vocab_emb):
			d = np.sqrt(np.sum((e-ve)**2))
			if d < mindist:
				mindist = d
				minword = v
		disemb.append(minword)
	return disemb

def plot_offsets(word1, word2, Q, K, V, vocab, vocab_emb):
    E = np.matmul(Q, K.T)/np.sqrt(3)
    evals = np.exp(E)
    A = evals/np.sum(evals)

    w1 = vocab_emb[vocab.index(word1)]
    w2 = vocab_emb[vocab.index(word2)]
    sorter = np.flip(np.argsort(A[9, :]))
    ovs = V[sorter]*np.repeat(A[9, sorter][:, np.newaxis], 3, axis=1)
    ovs = np.insert(ovs, 0, w1).reshape((-1, 3))
    starting_pts = np.cumsum(ovs, axis=0)
    xs, ys, zs = starting_pts[:, 0], starting_pts[:, 1], starting_pts[:, 2]
    As, Bs, Cs = ovs[1:, 0], ovs[1:, 1], ovs[1:, 2]
    
    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(w1[0], w1[1], w1[2])
    ax.scatter(w2[0], w2[1], w2[2])
    ax.quiver(xs[:-1], ys[:-1], zs[:-1], As, Bs, Cs)
    
    offset = np.sum(V*np.repeat(A[9, :][:, np.newaxis], 3, axis=1), axis=0)
    ax.quiver(xs[0], ys[0], zs[0], offset[0], offset[1], offset[2], color='r')
    ax.text(w1[0], w1[1], w1[2], word1, 'y')
    ax.text(w2[0], w2[1], w2[2], word2, 'y')
    ax.set_title('Value vector updates to one embedding')
    plt.show()

# ------- CAPTIONING -------

import json
import skimage.transform
from PIL import Image
from collections import Counter
from tqdm.notebook import tqdm

import torch
import torch.backends.cudnn as cudnn
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import torchvision
import torchvision.transforms as transforms
from torch.nn.utils.rnn import pack_padded_sequence

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class ImageEncoder(nn.Module):
	def __init__(self, enc_image_size=14):
		super(ImageEncoder, self).__init__()
		self.enc_image_size = enc_image_size

		# pretrained ImageNet ResNet-101
		resnet = torchvision.models.resnet101(pretrained=True)

		# remove linear and pool layers
		modules = list(resnet.children())[:-2]
		self.resnet = nn.Sequential(*modules)

		# resize image to fixed size using adaptive pool to allow input images of variable size
		self.adaptive_pool = nn.AdaptiveAvgPool2d((enc_image_size, enc_image_size))

		self.fine_tune()

	def fine_tune(self, fine_tune=True):
		"""
		Allow or prevent computation of the gradients for convolutional blocks 2 through 4 of the image encoder.
		:param fine_tune: boolean
		"""
		for param in self.resnet.parameters():
			param.requires_grad = False
		# if fine-tuning, fine-tune convolutional blocks 2 through 4
		for child in list(self.resnet.children())[5:]:
			for param in child.parameters():
				param.requires_grad = fine_tune

	def forward(self, images):
		"""
		Forward propagation.
		:param images: images, a tensor of dim (batch_size, 3, image_size, image_size)
		:return enc_images: encoded repr of images, a tensor of dim (batch_size, enc_image_size, enc_image_size, 2048)
		"""
		out = self.resnet(images)	   # (batch_size, 2048, image_size/32, image_size/32)
		out = self.adaptive_pool(out)   # (batch_size, 2048, enc_image_size, enc_image_size)
		out = out.permute(0, 2, 3, 1)   # (batch_size, enc_image_size, enc_image_size, 2048)
		return out

class BahdanauAttention(nn.Module):
	def __init__(self, encoder_dim, decoder_dim, attention_dim):
		"""
		:param encoder_dim: feature size of encoded images
		:param decoder_dim: size of decoder's RNN
		:param attention_dim: size of the attention network
		"""
		super(BahdanauAttention, self).__init__()
		self.encoder_attn = nn.Linear(encoder_dim, attention_dim)   # linear layer to transform encoder's output
		self.decoder_attn = nn.Linear(decoder_dim, attention_dim)   # linear layer to transform decoder's output
		self.full_attn = nn.Linear(attention_dim, 1)

	def forward(self, encoder_out, decoder_hidden):
		"""
		Forward propagation (uses Multiplicative attention).
		:param encoder_out: encoded images, a tensor of dim (batch_size, num_pixels, encoder_dim)
		:param decoder_hidden: previous decoder output, a tensor of dim (batch_size, decoder_dim)
		"""
		attn1 = self.encoder_attn(encoder_out)		  # (batch_size, num_pixels, attention_dim)
		attn2 = self.decoder_attn(decoder_hidden)	   # (batch_size, attention_dim)
		attn = self.full_attn(F.relu(attn1 + attn2.unsqueeze(1)))	# (batch_size, num_pixels, 1)

		# apply softmax to calculate weights for weighted encoding based on attention
		alpha = F.softmax(attn, dim=1)				  # (batch_size, num_pixels, 1)
		attn_weighted_encoding = (encoder_out * alpha).sum(dim=1)  # (batch_size, encoder_dim)
		alpha = alpha.squeeze(2)  # (batch_size, num_pixels)
		return attn_weighted_encoding, alpha

class DecoderWithAttention(nn.Module):
	def __init__(self, attention_dim, embed_dim, decoder_dim, vocab_size, encoder_dim=2048, dropout=0.5):
		"""
		:param attention_dim: size of attention network
		:param embed_dim: embedding_size
		:param decoder_dim: feature size of decoder's RNN
		:param vocab_size: size of vocabulary
		:param encoder_dim: feature size of encoded images
		:param dropout: dropout
		"""
		super(DecoderWithAttention, self).__init__()
		self.encoder_dim = encoder_dim
		self.attention_dim = attention_dim
		self.embed_dim = embed_dim
		self.decoder_dim = decoder_dim
		self.vocab_size = vocab_size
		self.dropout = dropout

		self.attention = BahdanauAttention(encoder_dim, decoder_dim, attention_dim)	 # attention network
		self.embedding = nn.Embedding(vocab_size, embed_dim)					# embedding layer
		self.dropout = nn.Dropout(p=dropout)

		self.decode_step = nn.LSTMCell(embed_dim + encoder_dim, decoder_dim, bias=True) # decoding LSTMCell
		self.init_h = nn.Linear(encoder_dim, decoder_dim)   # linear layer to find initial hidden state of LSTMCell
		self.init_c = nn.Linear(encoder_dim, decoder_dim)	# linear layer to find initial cell state of LSTMCell
		self.f_beta = nn.Linear(decoder_dim, encoder_dim)   # linear layer to create a sigmoid-activated gate

		self.fc = nn.Linear(decoder_dim, vocab_size)		# linear layer to find scores over vocabulary

		self.init_weights()	 # initialize some layers with the uniform distribution

	def init_weights(self):
		"""
		Initialize some layers with the uniform distribution for easier convergence.
		"""
		self.embedding.weight.data.uniform_(-0.1, 0.1)
		self.fc.bias.data.fill_(0)
		self.fc.weight.data.uniform_(-0.1, 0.1)

	def init_hidden_state(self, encoder_out):
		"""
		Creates the initial hidden and cell states for the decoder's LSTM based on the encoded images.
		:param encoder_out: encoded_images, a tensor of dimension (batch_size, num_pixels, encoder_dim)
		:return hidden_state, cell_state
		"""
		mean_encoder_out = encoder_out.mean(dim=1)  # (batch_size, encoder_dim)
		h = self.init_h(mean_encoder_out)   # (batch_size, decoder_dim)
		c = self.init_c(mean_encoder_out)   # (batch_size, decoder_dim)
		return h, c

	def forward(self, encoder_out, encoded_captions, caption_lens):
		"""
		Forward propagation.
		:param encoder_out: encoded images, a tensor of dimension (batch_size, enc_image_size, enc_image_size, encoder_dim)
		:param encoded_captions: encoded captions, a tensor of dimension (batch_size, max_caption_length)
		:param caption_lengths: caption lengths, a tensor of dimension (batch_size, 1)
		:return: scores for vocabulary, sorted encoded captions, decode lengths, weights, sort indices
		"""
		batch_size = encoder_out.size(0)

		# flatten image
		encoder_out = encoder_out.view(batch_size, -1, encoder_dim)	 # (batch_size, num_pixels, encoder_dim)
		num_pixels = encoder_out.size(1)

		# sort the input data by the decreasing caption length
		caption_lens, sort_idx = caption_lens.squeeze(1).sort(dim=0, descending=True)
		encoder_out = encoder_out[sort_idx]
		encoded_captions = encoded_captions[sort_idx]

		# embedding
		embeddings = self.embedding(encoded_captions)   # (batch_size, max_caption_length, embed_dim)

		# initialize lstm state
		h, c = self.init_hidden_state(encoder_out)	  # (batch_size, decoder_dim)

		# We won't decode at the <end> position, since we've finished generating as soon as we generate <end>
		# So, decoding lengths are caption lengths - 1
		decode_lens = (caption_lens - 1).tolist()

		# create tensors to hold word prediction scores and alphas
		predictions = torch.zeros(batch_size, max(decode_lens), vocab_size).to(device)
		alphas = torch.zeros(batch_size, max(decode_lens), num_pixels).to(device)

		# At each time-step, decode by attention-weighing the encoder's output based on the
		# decoder's previous hidden state output then generate a new word in the decoder with
		# the previous word and the attention weighted encoding
		for t in range(max(decode_lens)):
			# get the batch size for each time step t
			batch_size_t = sum([l > t for l in decode_lens])

			# get the attention weighted encodings (batch_size_t, encoder_dim)
			attention_weighted_encoding, alpha = self.attention(encoder_out[:batch_size_t], h[:batch_size_t])

			gate = F.sigmoid(self.f_beta(h[:batch_size_t]))	 # sigmoid gating scalar, (batch_size_t, encoder_dim)
			attention_weighted_encoding = gate * attention_weighted_encoding

			# get the decoder hidden state and cell state based on the embeddings of timestep t word
			# and the attention weighted encoding
			h, c = self.decode_step(
				torch.cat([embeddings[:batch_size_t, t, :], attention_weighted_encoding], dim=1),
				(h[:batch_size_t], c[:batch_size_t])
			)   # (batch_size_t, decoder_dim)

			# get the next word prediction
			preds = self.fc(self.dropout(h))	# (batch_size_t, vocab_size)

			# save the prediction and alpha for every time step
			predictions[:batch_size_t, t, :] = preds
			alphas[:batch_size_t, t, :] = alpha

		return predictions, encoded_captions, decode_lens, alphas, sort_idx


def generate_image_caption(encoder, decoder, image_path, word_map, beam_size=5):
    """
    Reads an image and captions it with beam search as well as plot attention maps.
    """
    k = beam_size

	# id to word mapping
    vocab_size = len(word_map)
    rev_word_map = {id: word for word, id in word_map.items()}

    
    img = torchvision.io.read_image(image_path)
    img = torchvision.transforms.Resize((256, 256))(img)
    img = img.numpy()
    
    if len(img.shape) == 3:
        if img.shape[0] != 3:
            img = img.repeat(3, 1, 1)
    else:
        raise AssertionError

    img = torch.FloatTensor(img / 255.)

    # encode the image
    encoder_out = encoder(img.unsqueeze(0).to(device))	 # (1, enc_image_size, enc_image_size, encoder_dim)
    enc_image_size = encoder_out.size(1)
    encoder_dim = encoder_out.size(3)

    # flatten encoded image representation
    encoder_out = encoder_out.view(1, -1, encoder_dim)  # (1, num_pixels, encoder_dim)
    num_pixels = encoder_out.size(1)

    # we'll treat the problem as having a batch size of k
    encoder_out = encoder_out.expand(k, num_pixels, encoder_dim)	# (k, num_pixels, encoder_dim)

    # tensor to store top k previous words at each step; now they're just <start>
    top_k_prev_words = torch.tensor([[word_map['<start>']]] * k, dtype=torch.long).to(device)   # (k, 1)

    # tensor to store top k sequences; now they're just <start>
    top_k_seqs = top_k_prev_words   # (k, 1)

    # tensor to store top k sequences' scores; now they're just 0
    top_k_scores = torch.zeros(k, 1).to(device)	 # (k, 1)

    # tensor to store top k sequences' alphas; now they're just 1s
    top_k_seqs_alpha = torch.ones(k, 1, enc_image_size, enc_image_size).to(device)  # (k, 1, enc_image_size, enc_image_size)

    # lists to store completed sequences along with their alphas and scores
    complete_seqs = []
    complete_seqs_alpha = []
    complete_seqs_scores = []

    # start decoding
    step = 1
    h, c = decoder.init_hidden_state(encoder_out)

    # s is a number less than or equal to k, because sequences are removed from this process once they hit <end>
    while True:
        embeddings = decoder.embedding(top_k_prev_words).squeeze(1)  # (s, embed_dim)

        attention_weighted_encoding, alpha = decoder.attention(encoder_out, h)  # (s, encoder_dim), (s, num_pixels, 1)
        alpha = alpha.view(-1, enc_image_size, enc_image_size)	 # (s, enc_image_size, enc_image_size)

        gate = F.sigmoid(decoder.f_beta(h))	  # gating scalar, (s, encoder_dim)
        attention_weighted_encoding = gate * attention_weighted_encoding

        h, c = decoder.decode_step(
            torch.cat([embeddings, attention_weighted_encoding], dim=1), (h, c)
        )   # (s, decoder_dim)

        scores = decoder.fc(h)	  # (s, vocab_size)
        scores = F.log_softmax(scores, dim=1)

        # add the scores to prev scores
        scores = top_k_scores.expand_as(scores) + scores	# (s, vocab_size)

        # all the k points will have the same score for the first step (since same k previous words, h, c)
        if step == 1:
            top_k_scores, top_k_words = scores[0].topk(k, 0, True, True)	# (s)
        else:
            # unroll and find top scores, and their unrolled indices
            top_k_scores, top_k_words = scores.view(-1).topk(k, 0, True, True)	# (s)

        # convert unrolled indices to actual indices of scores
        prev_word_inds = top_k_words // vocab_size  # (s)
        next_word_inds = top_k_words % vocab_size  # (s)

        # add new words to sequences, alphas
        top_k_seqs = torch.cat([top_k_seqs[prev_word_inds], next_word_inds.unsqueeze(1)], dim=1)	# (s, step + 1)
        top_k_seqs_alpha = torch.cat(
            [top_k_seqs_alpha[prev_word_inds], alpha[prev_word_inds].unsqueeze(1)], dim=1
        )   # (s, step + 1, enc_image_size, enc_image_size)

        # which sequences are incomplete (didn't reach <end>)?
        incomplete_inds = [ind for ind, next_word in enumerate(next_word_inds)
                            if next_word != word_map['<end>']]
        complete_inds = list(set(range(len(next_word_inds))) - set(incomplete_inds))

        # set aside complete sequences
        if len(complete_inds) > 0:
            complete_seqs.extend(top_k_seqs[complete_inds].tolist())
            complete_seqs_alpha.extend(top_k_seqs_alpha[complete_inds].tolist())
            complete_seqs_scores.extend(top_k_scores[complete_inds])
        k -= len(complete_inds)  # reduce beam length accordingly

        # proceed with incomplete sequences
        if k == 0:
            break

        top_k_seqs = top_k_seqs[incomplete_inds]
        top_k_seqs_alpha = top_k_seqs_alpha[incomplete_inds]
        h = h[prev_word_inds[incomplete_inds]]
        c = c[prev_word_inds[incomplete_inds]]
        encoder_out = encoder_out[prev_word_inds[incomplete_inds]]
        top_k_scores = top_k_scores[incomplete_inds].unsqueeze(1)
        top_k_prev_words = next_word_inds[incomplete_inds].unsqueeze(1)

        # break if things have been going on too long
        if step > 50:
            break
        step += 1

    # select sequence with max score
    i = complete_seqs_scores.index(max(complete_seqs_scores))
    seq = complete_seqs[i]
    alphas = complete_seqs_alpha[i]

    caption = [rev_word_map[ind] for ind in seq]

    img = Image.open(image_path).convert('RGB')
    img = img.resize([14 * 24, 14 * 24], Image.LANCZOS)

    # plot the image and attention map
    fig = plt.figure(figsize=(20, 8))
    for t in range(len(caption)):

        plt.subplot(int(np.ceil(len(caption) / 5.)), 5, t + 1)
        plt.text(0, 1, '%s' % (caption[t]), color='black', backgroundcolor='white', fontsize=12)
        plt.imshow(img)

        current_alpha = alphas[t]
        print(np.array(current_alpha).shape)
        alpha = skimage.transform.pyramid_expand(np.array(current_alpha), upscale=24, sigma=8)
        if t == 0:
            plt.imshow(alpha, alpha=0)
        else:
            plt.imshow(alpha, alpha=0.8)

        plt.set_cmap('gray')
        plt.axis('off')

    plt.show()
	
def continue_caption(encoder, decoder, image_path, word_map, part_caption, beam_size=5):
    part_caption = '<start> ' + part_caption + ' <end>'
    part_idxs = [word_map.get(w, word_map['<unk>']) for w in part_caption.split(' ')]
    
    
    k = beam_size
    vocab_size = len(word_map)
    # id to word mapping
    rev_word_map = {id: word for word, id in word_map.items()}
    
    img = torchvision.io.read_image(image_path)
    img = torchvision.transforms.Resize((256, 256))(img)
    img = img.numpy()
    
    if len(img.shape) == 3:
        if img.shape[0] != 3:
            img = img.repeat(3, 1, 1)
    else:
        raise AssertionError

    img = torch.FloatTensor(img / 255.)

    # encode the image
    encoder_out = encoder(img.unsqueeze(0).to(device))	 # (1, enc_image_size, enc_image_size, encoder_dim)
    enc_image_size = encoder_out.size(1)
    encoder_dim = encoder_out.size(3)

    # flatten encoded image representation
    encoder_out = encoder_out.view(1, -1, encoder_dim)  # (1, num_pixels, encoder_dim)
    num_pixels = encoder_out.size(1)

    # we'll treat the problem as having a batch size of k
    encoder_out = encoder_out.expand(k, num_pixels, encoder_dim)	# (k, num_pixels, encoder_dim)

    # tensor to store top k previous words at each step; now they're just <start>
    top_k_prev_words = torch.tensor([[word_map['<start>']]] * k, dtype=torch.long).to(device)   # (k, 1)

    # tensor to store top k sequences; now they're just <start>
    top_k_seqs = top_k_prev_words   # (k, 1)

    # tensor to store top k sequences' scores; now they're just 0
    top_k_scores = torch.zeros(k, 1).to(device)	 # (k, 1)

    # tensor to store top k sequences' alphas; now they're just 1s
    top_k_seqs_alpha = torch.ones(k, 1, enc_image_size, enc_image_size).to(device)  # (k, 1, enc_image_size, enc_image_size)

    # lists to store completed sequences along with their alphas and scores
    complete_seqs = []
    complete_seqs_alpha = []
    complete_seqs_scores = []

    # start decoding
    step = 1
    h, c = decoder.init_hidden_state(encoder_out)
    for idx in part_idxs:
        top_k_prev_words = torch.tensor([[idx]] * k, dtype=torch.long).to(device)  
        embeddings = decoder.embedding(top_k_prev_words).squeeze(1)  # (s, embed_dim)

        attention_weighted_encoding, alpha = decoder.attention(encoder_out, h)  # (s, encoder_dim), (s, num_pixels, 1)
        alpha = alpha.view(-1, enc_image_size, enc_image_size)	 # (s, enc_image_size, enc_image_size)

        gate = F.sigmoid(decoder.f_beta(h))	  # gating scalar, (s, encoder_dim)
        attention_weighted_encoding = gate * attention_weighted_encoding

        h, c = decoder.decode_step(
            torch.cat([embeddings, attention_weighted_encoding], dim=1), (h, c)
        )   # (s, decoder_dim)

        scores = decoder.fc(h)	  # (s, vocab_size)
        scores = F.log_softmax(scores, dim=1)

        # add the scores to prev scores
        scores = top_k_scores.expand_as(scores) + scores	# (s, vocab_size)
    
    top_k_prev_words = torch.tensor([[part_idxs[-1]]] * k, dtype=torch.long).to(device)
    embeddings = decoder.embedding(top_k_prev_words).squeeze(1)  # (s, embed_dim)

    attention_weighted_encoding, alpha = decoder.attention(encoder_out, h)  # (s, encoder_dim), (s, num_pixels, 1)
    alpha = alpha.view(-1, enc_image_size, enc_image_size)	 # (s, enc_image_size, enc_image_size)

    gate = F.sigmoid(decoder.f_beta(h))	  # gating scalar, (s, encoder_dim)
    attention_weighted_encoding = gate * attention_weighted_encoding

    h, c = decoder.decode_step(
        torch.cat([embeddings, attention_weighted_encoding], dim=1), (h, c)
    )   # (s, decoder_dim)

    scores = decoder.fc(h)	  # (s, vocab_size)
    scores = F.log_softmax(scores, dim=1)

    # add the scores to prev scores
    scores = top_k_scores.expand_as(scores) + scores	# (s, vocab_size)
    
    top_k_scores, top_k_words = scores[0].topk(k, 0, True, True)	# (s)
    
    prev_word_inds = top_k_words // vocab_size  # (s)
    next_word_inds = top_k_words % vocab_size  # (s)
    
    continuations = next_word_inds.unsqueeze(1)
    continuation_alphas = alpha[prev_word_inds].unsqueeze(1)

    img = Image.open(image_path).convert('RGB')
    img = img.resize([14 * 24, 14 * 24], Image.LANCZOS)

    fig = plt.figure()
    
    for t in range(k):
        plt.subplot(1, k, t + 1)
        plt.text(0, 1, '%s' % (rev_word_map[continuations[t].item()]), color='black', backgroundcolor='white', fontsize=12)
        plt.imshow(img)
        current_alpha = continuation_alphas[t].cpu().detach().numpy()
        alpha = skimage.transform.pyramid_expand(current_alpha[0, :, :], upscale=24, sigma=8)
        plt.imshow(alpha, alpha=0.8)

        plt.set_cmap('gray')
        plt.axis('off')
    fig.suptitle(' '.join([rev_word_map[ind] for ind in part_idxs[1:-1]]), fontsize=16)
    plt.show()

def prep_captioning():
    with open('tutorial_10_files/dataset_coco.json', 'r') as f:
    	data = json.load(f)

    max_len = 100			   # length of padded captions
    image_folder = '/content'   # folder with download images for train and val set
    min_word_freq = 5		   # words occuring less than this threshold will be labelled as <unk>
    captions_per_image = 5	  # number of captions per image
    word_freq = Counter()	   # Counter object to find the freq of each word
    
    # generate word map
    for img in data['images']:
    	for c in img['sentences']:
    		# update word frequency
    		word_freq.update(c['tokens'])
    
    # create word map (dict mapping words to ids with 0 reserved for padding)
    words = [word for word in word_freq.keys() if word_freq[word] > min_word_freq]
    word2id = {word: id for id, word in enumerate(words, 1)}
    word2id['<unk>'] = len(word2id) + 1
    word2id['<start>'] = len(word2id) + 1
    word2id['<end>'] = len(word2id) + 1
    word2id['<pad>'] = 0
    
    # create reverse word map (dict mapping ids to words)
    id2word = {value: word for word, value in word2id.items()}
    return word2id, id2word

# ------- ViTs -------
import torch
from matplotlib import pyplot as plt
import torch.nn.functional as F
from skimage.transform import resize
import numpy as np
import torchvision
from pathlib import Path
import requests


from tqdm.notebook import tqdm
import torch
import torch.nn.functional as F

class SelfAttention(nn.Module):
    """
    Canonical implementation of multi-head self attention.
    """

    def __init__(self, emb, heads=8, mask=False, kqnorm=False):
        """
        :param emb:
        :param heads:
        :param mask:
        :param kqnorm:
        """

        super().__init__()

        assert emb % heads == 0, f'Embedding dimension ({emb}) should be divisible by nr. of heads ({heads})'

        self.emb = emb
        self.heads = heads
        self.mask = mask

        s = emb // heads
        # - We will break the embedding into `heads` chunks and feed each to a different attention head

        self.tokeys    = nn.Linear(emb, emb, bias=False)
        self.toqueries = nn.Linear(emb, emb, bias=False)
        self.tovalues  = nn.Linear(emb, emb, bias=False)

        self.unifyheads = nn.Linear(emb, emb)

        self.kqnorm = kqnorm
        if kqnorm:
            self.kln = nn.LayerNorm([s])
            self.qln = nn.LayerNorm([s])
            
        self.attention_map = None

    def forward(self, x):

        b, t, e = x.size()
        h = self.heads
        assert e == self.emb, f'Input embedding dim ({e}) should match layer embedding dim ({self.emb})'

        s = e // h

        keys    = self.tokeys(x)
        queries = self.toqueries(x)
        values  = self.tovalues(x)

        keys    = keys.view(b, t, h, s)
        queries = queries.view(b, t, h, s)
        values  = values.view(b, t, h, s)

        if self.kqnorm:
            keys = self.kln(keys)
            queries = self.qln(queries)

        # -- We first compute the k/q/v's on the whole embedding vectors, and then split into the different heads.
        #    See the following video for an explanation: https://youtu.be/KmAISyVvE1Y

        # Compute scaled dot-product self-attention

        # - fold heads into the batch dimension
        keys = keys.transpose(1, 2).contiguous().view(b * h, t, s)
        queries = queries.transpose(1, 2).contiguous().view(b * h, t, s)
        values = values.transpose(1, 2).contiguous().view(b * h, t, s)

        queries = queries / (e ** (1/4))
        keys    = keys / (e ** (1/4))
        # - Instead of dividing the dot products by sqrt(e), we scale the keys and values.
        #   This should be more memory efficient

        # - get dot product of queries and keys, and scale
        dot = torch.bmm(queries, keys.transpose(1, 2))

        assert dot.size() == (b*h, t, t)

        if self.mask: # mask out the upper half of the dot matrix, excluding the diagonal
            mask_(dot, maskval=float('-inf'), mask_diagonal=False)
                    
        dot = F.softmax(dot, dim=2)
        # - dot now has row-wise self-attention probabilities
        
        # save attention map 
        self.attention_map = dot
        
        # apply the self attention to the values
        dot = torch.bmm(dot, values)      
        out = dot.view(b, h, t, s)

        # swap h, t back, unify heads
        out = out.transpose(1, 2).contiguous().view(b, t, s * h)

        return self.unifyheads(out)


def download_OxfordPetIIIt_model(name='model_oxford.pt'):
    
    source_url = 'https://is.muni.cz/el/fi/jaro2023/PA228/um/exercises/tutorial_10/model_oxford.pt'
    model_oxford_path = Path(name)

    if (not model_oxford_path.exists()):
        print(f'\tdownloading ... {source_url}')
        content = requests.get(source_url).content
        model_oxford_path.open("wb").write(content)
        
    return str(model_oxford_path)

def dataset_OxfordPetIIIt():
    data_path = Path('data')

    config = {
        'name' : 'OxfordPetIIIt',
        'batch_size' : 32,
        'image_size' : 256,
        'patch_size' : 16,
        'num_classes': 37,
        'channels'   : 3,
        'dim' : 512,
        'depth' : 4,
        'heads' : 16,
        'mlp_dim': 64,
        'lr': 0.0001
    }

    transform_oxford_train = torchvision.transforms.Compose(
                [torchvision.transforms.ToTensor(),
                 torchvision.transforms.RandomHorizontalFlip(p=0.5),
                 torchvision.transforms.RandomRotation(degrees=(-20, 20)),
                 torchvision.transforms.Resize(int(config['image_size']*1.2)),
                 torchvision.transforms.RandomResizedCrop(scale=(0.7, 1.0), size=(config['image_size'], config['image_size'])),
                 torchvision.transforms.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225))])
    
    transform_oxford = torchvision.transforms.Compose(
            [torchvision.transforms.ToTensor(),
             torchvision.transforms.Resize((config['image_size'], config['image_size'])),
             torchvision.transforms.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225))])
    
    torchvision.transforms.RandomHorizontalFlip(p=0.5)

    train_ds = torchvision.datasets.OxfordIIITPet(data_path,
                                          split='trainval',
                                          download=True,
                                          transform=transform_oxford_train)
    valid_ds = torchvision.datasets.OxfordIIITPet(data_path,
                                          split='test',
                                          download=True,
                                          transform=transform_oxford)

    train_dl = torch.utils.data.DataLoader(train_ds, batch_size=config['batch_size'], shuffle=True)
    valid_dl = torch.utils.data.DataLoader(valid_ds, batch_size=config['batch_size'], shuffle=True)
    
    return (train_dl, valid_dl), config

def normalize_inverse(img):
    
    if img.shape[-1] == 3:
    
        mean = np.array([0.485, 0.456, 0.406]).reshape(1, 1, 3)
        std = np.array([0.229, 0.224, 0.225]).reshape(1, 1, 3)
    else:
        mean = np.array([0.1307,]).reshape(1, 1, 1)
        std = np.array([0.3081,]).reshape(1, 1, 1)
        
        img.reshape((img.shape[0], -1, 1))
        
    
    img = (img * std + mean) * 255
    return img.astype(int)


def overlay_mask(img, mask, alpha=.5):
    return ((plt.cm.coolwarm(mask)[:, :, :3] * 255) * alpha + img * (1 - alpha)).astype(np.uint8)


def plot_in_grid(samples, n_cols=6):
    
    n_rows = len(samples) // n_cols
    
    rows = []
    for row_idx in range(n_rows):
        rows.append(np.concatenate(samples[row_idx*n_cols:(row_idx + 1)*n_cols], axis=1))
    grid = np.concatenate(rows, axis=0)
    
    plt.figure(figsize=(12, 12))
    plt.imshow(grid)
    plt.show()
    
    
def decode_mask(att, sample_idx, head_idx, config):
    '''
    decodes an attention map into a heatmap, where patches of high value
    are considered more important for the classification
    '''
        
    # take he first column vector (cls_token) of attention map
    attention_vector = att[sample_idx + head_idx * config['batch_size'], 0]
    mask = attention_vector.detach().cpu().numpy()

    # remove the first value from the vector
    mask = mask[1:]
    
    # reshape vector to 2D image
    n_patches = config['image_size'] // config['patch_size']
    mask = mask.reshape((n_patches, n_patches))
    mask = (mask * 255).astype(int)
    mask = resize(mask, (config['image_size'], config['image_size']), order=0)
    
    return mask


def plot_attention(model, dl, config, dev='cpu'):
    # get one batch
    for xb, yb in dl:
        break

    # get attention maps
    model.eval()
    att_maps = model.get_attention_maps(xb.to(dev))
    n_samples = 32

    samples = []

    # iterate over samples in a batch
    for sample_idx in range(n_samples):

        label = yb[sample_idx].detach().cpu().numpy()
        img = xb[sample_idx].detach().cpu().numpy().transpose(1, 2, 0)
        img = normalize_inverse(img)

        # take only the last attention map
        att = att_maps[-1]

        bxh, t1, t2 = att.shape
        assert bxh == config['batch_size'] * config['heads']
        assert t1 == t2

        # iterate over heads
        mask_list = []
        for head_idx in range(config['heads']):
            mask = decode_mask(att, sample_idx, head_idx, config)
            mask_list.append(mask)

        # reduce masks over a heads
        mean_mask = np.mean(mask_list, axis=0).astype(np.uint8)
        if mean_mask.max() > 0:
            mean_mask = mean_mask / mean_mask.max() 

        # overlay the input image
        overlay = overlay_mask(img, mean_mask)
        samples.append(overlay)

    plot_in_grid(samples, n_cols=8)